#python

#
# Title: Area Weighted Normals
# Version: 2.3
# Author: James O'Hare
# http://www.farfarer.com
# jamesohare@gmail.com
#

import lx
from farfarer_vnorm import vnormkit
from math import acos, degrees

#import time
#t1 = time.time()

def stepmon():
	m.step()

# Grab the version check result.
slowVersion = vnormkit.isSP1()

# Grab the settings for the script from user values.
highQualityNormals = lx.eval1("user.value vnormkit.areahqnorm ?")
respectSmoothGroups = lx.eval1("user.value vnormkit.areagroups ?")
respectEdgeSmoothing = lx.eval1("user.value vnormkit.areaedges ?")
useCustomNormalMapName = lx.eval1("user.value vnormkit.usenormname ?")
customNormalMapName = lx.eval1("user.value vnormkit.normname ?")
areaWeighted = lx.eval1("user.value vnormkit.areaweighted ?")
angleWeighted = lx.eval1("user.value vnormkit.angleweighted ?")
fullWeight = lx.eval1("user.value vnormkit.areafullweight ?")

# Save current selection, for restoring later.
selectionType = vnormkit.getSelectionType()
selRestore = vnormkit.saveSelection()

# Ensure vertex normal maps all share the same name.
vnormkit.renameNormalMapsDefault()

s = lx.Service('layerservice')
s.select('layers', 'main')
mainLayer = s.query1('layers')
mainLayerID = s.query1('layer.ID')

s.select('layers', 'fg')
layers = s.queryN('layers')

# Drop selections.
lx.eval('select.drop vertex')
lx.eval('select.drop edge')
lx.eval('select.drop polygon')
lx.eval('select.drop item')

for layer in layers:
	layerstr = str(layer)
	s.select('layer.index', layerstr )
	layerID = s.query1('layer.id')
	lx.eval('select.subItem %s set mesh' % layerID)
	lx.eval('select.type polygon')

	if lx.eval1('query layerservice poly.N ? all') > 0:

		#________________________________________________________________________________________________________________________ Bake Normals
		if vnormkit.doesNormalMapExist(vnormkit.defaultNormalName):
			lx.eval('!!select.vertexMap %s norm 3' % vnormkit.defaultNormalName)
		lx.eval('vertMap.normals')

		#________________________________________________________________________________________________________________________ Store Open Edges
		if respectEdgeSmoothing or respectSmoothGroups:
			vnormkit.saveOpenEdges()

		#________________________________________________________________________________________________________________________ Split Edge Smoothing

		if respectEdgeSmoothing:
			# Select and split the hard edges.
			lx.eval('select.type edge')
			lx.eval('select.drop edge')
			if vnormkit.doesEdgeSetExist(vnormkit.hardEdgeSetName):
			  	lx.eval('select.useSet %s select' % vnormkit.hardEdgeSetName)
			  	if lx.eval1('query layerservice edge.N ? selected') > 0:
			  		lx.eval('edge.split')

		#________________________________________________________________________________________________________________________ Split Smoothing Groups

		if respectSmoothGroups:
			smoothingGroups = lx.evalN('query sceneservice tags.list ? SMGP')

			# Select the boundaries of each smoothing group.
			lx.eval('select.drop edge')
			for smoothingGroup in smoothingGroups:
				lx.eval('select.type polygon')
				lx.eval('select.drop polygon')
				lx.eval('select.polygon add smoothGroup face %s' % smoothingGroup)
				# Split them if there are any boundaries.
				if lx.eval1('query layerservice poly.N ? selected') > 0:
					lx.eval('select.boundary')
					lx.eval('edge.split')

		#________________________________________________________________________________________________________________________ Monitor and Vert/Poly Lists

		s.select('polys', 'all')
		polygons = s.queryN('polys')
		numPolys = len(polygons)
		s.select('verts', 'all')
		vertices = s.queryN('verts')
		numVerts = len(vertices)

		# Set up monitor for progress bar.
		m = lx.Monitor()
		monitorLength = numPolys + numVerts + 1
		m.init(monitorLength)

		#________________________________________________________________________________________________________________________ Polygon Normals & Areas
		# Create an multi-dim list to hold the normals of each polygon.
		polygonNormals = [[0.0] * 3 for i in range(numPolys)]

		# Create an array to hold the areas of each polygon.
		if areaWeighted or fullWeight:
			# Only initialise it fully if we need it.
			polygonAreas = [0.0 for i in range(numPolys)]

		for poly in polygons:
			s.select('poly.index', poly)
			# Get vertices of polygon, count them and find out triangle count.
			polyVerts = s.queryN('poly.vertList')
			vertCount = len(polyVerts)
			triCount = vertCount - 2

			# Create empty list to store polygon vertex positions.
			# Faster if area/full/hqnorm is active as it only needs to be accessed once.
			polyVPos = []
			if highQualityNormals or areaWeighted or fullWeight:
				for i in range(vertCount):
					s.select('vert.index', str(polyVerts[i]))
					vPos = s.queryN('vert.pos')
					polyVPos.append(vPos)

			# Create empty list for polygon normal.
			polyNormal = []

			if highQualityNormals:
				# Let's do this shit properly, average the normals of all of the alternate pairs of edges for this polygon.
				# More accurate result, but slower to calculate.
				# Set the normal to zero to start with.
				polyNormal = [0.0] * 3

				for i in range(vertCount):
					# Vert positions.
					next = (i + 1) % vertCount

					currentPos = polyVPos[i]
					nextPos = polyVPos[next]

					# Poly normal.
					currentMinusNext = vnormkit.vector_subtract(currentPos, nextPos)
					currentPlusNext = vnormkit.vector_add(currentPos, nextPos)

					polyNormal[0] += currentMinusNext[1] * currentPlusNext[2]
					polyNormal[1] += currentMinusNext[2] * currentPlusNext[0]
					polyNormal[2] += currentMinusNext[0] * currentPlusNext[1]

				# Average the normals of the polygon's triangles by normalizing the result.
				polyNormal = vnormkit.vector_normalize(polyNormal)
			else:
				# Use modo's polygon normal. Not truly accurate, but faster.
				s.select('poly.index', poly)
				polyNormal = s.queryN('poly.normal')

			# Store values in the arrays.
			polygonNormals[int(poly)] = polyNormal

			if areaWeighted or fullWeight:
					# Calculate the area of the polygon by adding the area of it's triangles.
					polyArea = 0.0
					vert1Pos = polyVPos[0]
					for i in range(triCount):
						vert2Pos = polyVPos[i+1]
						vert3Pos = polyVPos[i+2]
						triArea = vnormkit.triangle_area(vert1Pos, vert2Pos, vert3Pos)
						#lx.out('Area of triangle for %s is %s' % (poly, triArea))
						polyArea += triArea
					#lx.out('Area of polygon %s is %s' % (poly, polyArea))
					polygonAreas[int(poly)] = polyArea

			stepmon()
			# lx.out('For poly %s, area is %s, normal is %s' % (poly, polyArea, polyNormal))

		#________________________________________________________________________________________________________________________ Vertex Normals & Angles

		lx.eval('select.vertexMap %s norm replace' % vnormkit.defaultNormalName)
		lx.eval('select.drop vertex')

		for vert in vertices:
			# Create empty dictionary to store the angles of edges where polygons share a vertex.
			# angleDict[vertIndex][polyIndex] = angle of edges at that polygon for that vertex (in radians)
			angleDict = {}

			s.select('vert.index', vert)
			vertPolygons = s.queryN('vert.polyList')

			# Calculate vertex angles.
			if angleWeighted:
				vertVerts = s.queryN('vert.vertList')
				vert0Pos = s.queryN('vert.pos')

				for vertPoly in vertPolygons:
					s.select('poly.index', str(vertPoly))
					polyVerts = s.queryN('poly.vertList')

					vertList = [i for i in polyVerts if i in vertVerts]

					s.select('vert.index', str(vertList[0]))
					vert1Pos = s.queryN('vert.pos')
					s.select('vert.index', str(vertList[1]))
					vert2Pos = s.queryN('vert.pos')

					edge1 = vnormkit.vector_normalize(vnormkit.vector_subtract(vert0Pos, vert1Pos))
					edge2 = vnormkit.vector_normalize(vnormkit.vector_subtract(vert0Pos, vert2Pos))
					dp = vnormkit.vector_dot(edge1, edge2)
					angle = acos(dp)

					angleDict[vertPoly] = angle

					#lx.out('Vertex %s for Poly %s has \n  Edge IDs ((%s,%s) (%s,%s)) \n  Positions (%s,%s,%s) \n  Edge vectors (%s,%s) \n  With angle %s between them.' % (vert, vertPoly, vert, vertList[0], vert, vertList[1], vert0Pos, vert1Pos, vert2Pos, edge1, edge2, angle))

			# Calculate vertex normals.

			# Create a zero vector to start with.
			vertNormal = [0.0] * 3

			if areaWeighted:
				# Get normals of all the polygons the vertex belongs to, multiply them by polygon area and add them together.
				for vertPoly in vertPolygons:
					multiplier = polygonAreas[vertPoly]
					if angleWeighted:
						multiplier *= angleDict[vertPoly] # Multiply by angle.
					polyNormal = vnormkit.vector_multiply_scalar(polygonNormals[vertPoly], multiplier) # Multiply by area.
					vertNormal = vnormkit.vector_add(vertNormal, polyNormal)
			elif fullWeight:
				# Get areas of all the polygons the vertex belongs to and pick the normal of the largest polygon.
				largestArea = 0.0
				for vertPoly in vertPolygons:
					polyArea = polygonAreas[vertPoly]
					if polyArea > largestArea:
						vertNormal = polygonNormals[vertPoly]
						largestArea = polyArea
			else:
				# Get normals of all the polygons the vertex belongs to and add them together.
				for vertPoly in vertPolygons:
					polyNormal = polygonNormals[vertPoly]
					if angleWeighted:
						polyNormal = vnormkit.vector_multiply_scalar(polyNormal, angleDict[vertPoly]) # Multiply by area.
					vertNormal = vnormkit.vector_add(vertNormal, polyNormal)

			# Normalize the new vertex normal.
			vertNormal = vnormkit.vector_normalize(vertNormal)

			# Put them into the vertex normal map.
			if slowVersion:
				# If we're using 701 SP1, we have to use a slow-ass workaround for broken vertMap.setVertex.
				lx.eval('select.element %s vertex set %s' % (layer, vert))
				lx.eval('vertMap.setValue normal 0 %s' % vertNormal[0])
				lx.eval('vertMap.setValue normal 1 %s' % vertNormal[1])
				lx.eval('vertMap.setValue normal 2 %s' % vertNormal[2])
			else:
				# If we're not using 701 SP1.
				# This way is waaaaaay faster.
				lx.eval('vertMap.setVertex %s normal 0 %s %s' % (vnormkit.defaultNormalName, vert, vertNormal[0]))
				lx.eval('vertMap.setVertex %s normal 1 %s %s' % (vnormkit.defaultNormalName, vert, vertNormal[1]))
				lx.eval('vertMap.setVertex %s normal 2 %s %s' % (vnormkit.defaultNormalName, vert, vertNormal[2]))

			if slowVersion:
				lx.eval('select.drop vertex')

			stepmon()

		#________________________________________________________________________________________________________________________ Weld Verts & Re-Split Edges

		if respectEdgeSmoothing or respectSmoothGroups:
			vnormkit.reweldAndSplit()
			
			# Drop selection.
			lx.eval('select.drop vertex')
			lx.eval('select.drop edge')
			lx.eval('select.drop polygon')

		stepmon() # Last step.

# Restore original layer selection.
lx.eval('select.subItem %s set mesh' % mainLayerID)
for layer in layers:
	layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
	lx.eval('select.subItem %s add mesh' % layerID)

# Ensure vertex normal maps all share the same name.
vnormkit.renameNormalMapsCustom(True)

# Restore original selection & selection type.
vnormkit.restoreSelection(selRestore)
lx.eval('select.type %s' % selectionType)

#t2 = time.time()
#timeTaken = t2 - t1
#lx.out('Took %s' % timeTaken)