#python
#
# Title: Edge Smoothing
# Version: 1.5
# Author: James O'Hare
# http://www.farfarer.com
# jamesohare@gmail.com
#

#
# Description:
# Tools to soften and harden edges of a mesh and the vertex normals stored into a vertex normal map.
# The edges to be made hard are stored in an edge pick map.
# Also contains helper functions for setting up vertex normal map, smoothing materials (for baking).
#

#
# Note:
# When 'smoothing' an edge, it will bake the smoothing that's already applied to the mesh.
# This includes the material's smoothing properties, which is why it is important to ensure the material is set to
# smooth via the 'smoothing angle' being set to 180 and the 'smoothing amount' being set to 100%.
# This script contains a tool that will set this up for you on the materials applied to your mesh.
#

#
# Installation:
# Put this .py file into your Scripts directory.
# Put the .cfg file into your Configs directory to add a form of tools.
#
# Open the Form Editor from System > Form Editor
# Go to your Edge tab and then use the Find Form button from the editor and click on the grey space in the edge menu.
# In the Form Editor, open Groups > (ungrouped) and find the form labeled Edge Smoothing.
# Click and drag the Edge Smoothing form to the bottom of the Edge form you found earlier.
# And you're done :)
#
# The commands are listed below if you don't want to install the form.
#

#
# Usage:
# @edgesmooth.py arguments
# 
# Arguments			- Function
#
# setup				- Creates/renames vertex normal maps, smooths the materials applied to the mesh, bakes normals.
#
# harden			- Hardens selected edges. Has no effect if no edges selected.
# harden all		- Hardens all edges, regardless of selection.
# harden polyborder	- Hardens the boundary edges of the selected polygons.
# harden uv			- Hardens all of the UV borders of the currently selected UV map.
# harden angle		- Hardens edges whose polygons are at an angle specified by the user value "edgesmoothAngle".
#
# soften			- Softens selected edges. Has no effect if no edges selected.
# soften all		- Softens all edges, regardless of selection.
# soften polyborder	- Softens the boundary edges of the selected polygons.
#
# select hard		- Selects the edges marked as hard.
# select soft		- Selects the edges marked as soft (or rather, the edges not marked as hard).
#
# update			- Rebakes the smoothing information. Useful if you've made any geometry edits to the mesh.
# toggle			- Deletes or recreates the vertex normal map. Does not remove the hard edge selection set.
# smoothmat			- Simply smooths all the materials applied to the mesh.
#

#
# Updates:
#
# 1.5
# Switched to Python.
# Added command to create hard edges from material smoothing angle.
# Added command to create hard edges from arbitrary smoothing angle.
#
# 1.4
# Fixed selection of soft/hard edges not working.
#
# 1.3
# Works across all selected layers.
# Adds the option to soften/harden the border edges of selected polygons.
#
# 1.2
# Restores original selection type now (would previously dump you in edge mode no matter what mode you were in before).
# Will do nothing if you try and do run the script outside of vert/edge/poly mode.
# Updated script for 701, should be faster and more reliable in 701 as it uses the new edge.splitNormals tool.
#
# 1.1
# Added option to harden UV borders of selected UV map.
# @edgesmooth.pl harden uv
#

import lx
from farfarer_vnorm import vnormkit
from math import cos, acos, radians, degrees, fabs

########################################################################################################################
# Helper functions.
########################################################################################################################

# Store boundary edges as a selection set.
# Return whether there are any.
def saveOpenEdges():
	lx.eval('select.drop edge')
	lx.eval('select.boundary')
	if lx.eval1('query layerservice edge.N ? selected') > 0:
		#lx.out('Open edges.')
		lx.eval('select.editSet %s add \"\"' % vnormkit.openEdgeSetName)
	lx.eval('select.drop edge')

# Drop the current selection.
def dropSelection():
	lx.eval('select.drop vertex')
	lx.eval('select.drop edge')
	lx.eval('select.drop polygon')

# Update the vertex normal maps for modo 701. Much easier to do...
#... wipe the current vertex normal map, then bake a new one and use edge.splitNormals on the hard edges.
def doEdgeSmoothing701():
	lx.eval('vertMap.normals')

	# Pull up the hard edge selection set, split the normals for those edges, drop the selection.
	lx.eval('select.type edge')
	#lx.out('Attempting to select hard edge set.')
	# try:
	# 	lx.eval('!!select.vertexMap %s epck replace' % vnormkit.hardEdgeSetName)
	# except RuntimeError:
	# 	#lx.out('Failed to select edge set - you don\'t have any hard edges.')
	# 	lx.eval('noop')
	# else:
	# 	#lx.out('Succeeded to select edge set.')
	# 	lx.eval('!!select.useSet %s select' % vnormkit.hardEdgeSetName)
	# 	if lx.eval1('query layerservice edge.N ? selected') > 0:
	# 		lx.eval('edge.splitNormals %s' % vnormkit.defaultNormalName)

	if vnormkit.doesEdgeSetExist(vnormkit.hardEdgeSetName):
		lx.eval('!!select.useSet %s select' % vnormkit.hardEdgeSetName)
	 	if lx.eval1('query layerservice edge.N ? selected') > 0:
	 		lx.eval('edge.splitNormals %s' % vnormkit.defaultNormalName)

# Update the vertex normal maps for modo 601. Split hard edges, baking normals, and then welding edges back together...
# ...being sure not to weld together edges that were previously already open.
def doEdgeSmoothing601():
	vnormkit.saveOpenEdges()

	lx.out('Moose 1');

	# Split selected edges.
	#lx.out('Attempting to select hard edge set.')
	#try:
	#	lx.eval('!!select.useSet %s select' % vnormkit.hardEdgeSetName)
	#except RuntimeError:
	#	#lx.out('Failed to select edge set - you don\'t have any hard edges.')
	#	lx.eval('noop')
	#else:
	#	#lx.out('Succeeded to select edge set.')
	#	#lx.out('Splitting edges.')
	#	if lx.eval1('query layerservice edge.N ? selected') > 0:
	#		lx.eval('edge.split')

	if vnormkit.doesEdgeSetExist(vnormkit.hardEdgeSetName):
		lx.eval('!!select.useSet %s select' % vnormkit.hardEdgeSetName)
	 	if lx.eval1('query layerservice edge.N ? selected') > 0:
	 		lx.eval('edge.split')

	# Bake vertex normals.
	lx.eval('vertMap.normals')

	# Weld all open edges.
	lx.eval('select.type edge')
	lx.eval('select.drop edge')
	lx.eval('select.boundary')
	lx.eval('select.convert vertex')
	lx.eval('tool.set vert.merge on')
	lx.eval('tool.setAttr vert.merge dist 0.0')
	lx.eval('tool.doApply')
	lx.eval('tool.set vert.merge flush 0')
	lx.eval('select.drop vertex')

	lx.out('Moose 2');

	vnormkit.reweldAndSplit()
	
	# Drop selection.
	lx.eval('select.drop vertex')
	lx.eval('select.drop edge')
	lx.eval('select.drop polygon')

	lx.out('Moose 3');

	# Drop selection.
	dropSelection()

	lx.out('Moose 4');

# Update the vertex normal maps depending on the modo version.
def doEdgeSmoothing():
	if lx.eval1('query platformservice appversion ?') < 701:
		#lx.out('Pre-701.')
		doEdgeSmoothing601()
	else:
		#lx.out('Post-701.')
		doEdgeSmoothing701()

# Update the smoothing for each layer.
def updateSmoothing():
	lx.eval('select.type edge')
	lx.eval('select.drop edge')

	# Get a list of the active layers.
	layers = lx.evalN('query layerservice layers ? fg')

	# Grab the main layer ID. We'll select this first later on to ensure that it remains the main layer.
	mainLayer = lx.eval1('query layerservice layers ? main')
	mainLayerID = lx.eval1('query layerservice layer.id ? %s' % mainLayer)

	# Select each layer. Do edge smoothing on it.
	for layer in layers:
		#Select only one layer.
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		layerName = lx.eval1('query layerservice layer.name ? %s' % layer)
		lx.eval('select.subItem %s set mesh' % layerID)
		#lx.out('Selected %s.' % layerID)
		doEdgeSmoothing()

	# Restore original layer selection. Main layer first.
	lx.eval('select.subItem %s set mesh' % mainLayerID)
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		lx.eval('select.subItem %s add mesh' % layerID)

# Set up a normal map if one doesn't exist, rename it if it does.
def setupNormalMap():
	# Get a list of the active layers.
	layers = lx.evalN('query layerservice layers ? fg')

	# Grab the main layer ID. We'll select this first later on to ensure that it remains the main layer.
	mainLayer = lx.eval1('query layerservice layers ? main')
	mainLayerID = lx.eval1('query layerservice layer.id ? %s' % mainLayer)

	# Select each layer. Bake vertex map.
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		lx.eval('select.subItem %s set mesh' % layerID)
		#lx.out('Selecting %s.' % layerID)
		try:
			lx.eval('!!select.vertexMap %s norm replace' % vnormkit.defaultNormalName)
		except RuntimeError:
			lx.eval('vertMap.normals')

	# Restore original layer selection. Main layer first.
	lx.eval('select.subItem %s set mesh' % mainLayerID)
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		lx.eval('select.subItem %s add mesh' % layerID)

# Hack to update the smoothing in the scene. Needs a dialog popup to force modo to refresh the viewport and thus the shading.
def smoothMaterialsDialog():
	lx.eval('dialog.setup info')
	lx.eval('dialog.title {Material smoothing altered.}')
	lx.eval('dialog.msg {The materials have been made smooth.\nI need this dialog here or modo won\'t update the smoothing to bake it.\nSorry.}')
	lx.eval('dialog.result ok')
	lx.eval('dialog.open')

# Set all the materials in the scene to totally smooth.
def smoothMaterials():
	#lx.out('Smoothing materials.')
	needsRefreshing = False
	mats = lx.evalN('query layerservice materials ?')
	for mat in mats:
		matName = lx.eval1('query layerservice material.id ? %s' % mat)
		#lx.out('Material ID: %s' % matName)
		#lx.out('Material ID Type: %s' % type(matName))
		if matName is not None:
			lx.eval('select.subItem %s set textureLayer' % matName)
			if lx.eval1('item.channel advancedMaterial$smooth ?') != 1.0:
				lx.eval('item.channel advancedMaterial$smooth 1.0')
				needsRefreshing = True
			if lx.eval1('item.channel advancedMaterial$smAngle ?') != radians(180.0):
				lx.eval('item.channel advancedMaterial$smAngle 180.0')
				needsRefreshing = True
	if needsRefreshing:
		smoothMaterialsDialog()

def hardEdgesByAngle(angle):
	layers = lx.evalN('query layerservice layers ? fg')
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		edges = lx.evalN('query layerservice edges ? all')
		for edge in edges:
			edgeVerts = lx.evalN('query layerservice edge.vertList ? %s' % edge)
			edgePolys = lx.evalN('query layerservice edge.polyList ? %s' % edge)
			if len(edgePolys) > 1:
				normalA = lx.evalN('query layerservice poly.normal ? %s' % edgePolys[0])
				normalB = lx.evalN('query layerservice poly.normal ? %s' % edgePolys[1])
				# DONE IN DEGREES
				dotProd = vnormkit.vector_dot(normalA, normalB)
				dp_angle = degrees(acos(dotProd))
				#lx.out('Angle: %s DotProdAngle: %s' % (angle, dp_angle))

				if dp_angle > angle:
					lx.eval('select.element %s edge set %s %s' % (layer, edgeVerts[0], edgeVerts[1]))
					lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)

def hardEdgesByMaterial():

	materialNameToAngle = {}

	mats = lx.evalN('query layerservice materials ?')
	for mat in mats:
		matID = lx.eval1('query layerservice material.id ? %s' % mat)
		matName = lx.eval1('query layerservice material.name ? %s' % mat)
		lx.eval('select.subItem %s set textureLayer' % matID)
		smoothAngle = lx.eval1('item.channel advancedMaterial$smAngle ?')
		#lx.out('Material %s has angle %s (%s)' % (matName, smoothAngle, degrees(smoothAngle)))
		materialNameToAngle[matName] = smoothAngle

	layers = lx.evalN('query layerservice layers ? fg')
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.id ? %s' % layer)
		edges = lx.evalN('query layerservice edges ? all')
		for edge in edges:
			edgeVerts = lx.evalN('query layerservice edge.vertList ? %s' % edge)
			edgePolys = lx.evalN('query layerservice edge.polyList ? %s' % edge)
			normalA = lx.evalN('query layerservice poly.normal ? %s' % edgePolys[0])
			normalB = lx.evalN('query layerservice poly.normal ? %s' % edgePolys[1])
			dotProd = vnormkit.vector_dot(normalA, normalB)
			# DONE IN RADIANS
			dp_angle = acos(dotProd)

			# Get the largest angle of the polys' materials, use that.
			# It's ridiculous, but that's the way modo does it.
			angle = 0.0
			for poly in edgePolys:
				polyTags = lx.evalN('query layerservice poly.tags ? %s' % poly)
				polyTagTypes = lx.evalN('query layerservice poly.tagTypes ? %s' % poly)
				for tag, tagType in zip(polyTags, polyTagTypes):
					if (tagType == 'MATR'):
						polyAngle = materialNameToAngle[tag]
						if polyAngle > angle:
							angle = polyAngle

			if dp_angle > angle:
				lx.eval('select.element %s edge set %s %s' % (layer, edgeVerts[0], edgeVerts[1]))
				lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)

def checkUVSelection():
	selectedUVMapName = lx.eval1('vertMap.list type:txuv ?')
	if selectedUVMapName == '_____n_o_n_e_____':
		lx.eval('dialog.setup error')
		lx.eval('dialog.title {No UV Map Selected.}')
		lx.eval('dialog.msg {You must have a UV map selected to harden it\'s borders.}')
		lx.eval('dialog.result ok')
		lx.eval('dialog.open')
		exit(1)

########################################################################################################################
# The script itself.
########################################################################################################################
# Store original selection type.
selectionType = vnormkit.getSelectionType()
#lx.out('Selection type is %s' % selectionType)
selRestore = vnormkit.saveSelection()
vnormkit.renameNormalMapsDefault()

lx.eval('select.type edge')

args = lx.args()
# #______________________________________________________________________________________________________ Setup
# if args[0].lower() == 'setup':
# 	smoothMaterials()
# 	updateSmoothing()
#______________________________________________________________________________________________________ Smooth Materials
if args[0].lower() == 'smoothmat':
	smoothMaterials()
#______________________________________________________________________________________________________ Toggle Smoothing
elif args[0].lower() == 'toggle':
	if lx.eval1('query layerservice vmap.N ? normal') > 0:
		# If normal map exists, delete it.
		normalMap = lx.eval1('query layerservice vmaps ? normal')
		normalMapName = lx.eval1('query layerservice vmap.name ? %s' % normalMap)
		lx.eval('select.vertexMap %s norm replace' % normalMapName)
		lx.eval('!!vertMap.delete norm')
		#lx.out('Disabled.')
	else:
		# If normal map doesn't exist, create it, run smoothing.
		smoothMaterials()
		updateSmoothing()
		#lx.out('Enabled.')
	dropSelection()
#________________________________________________________________________________________________________ Edge Selection
elif args[0].lower() == 'select':
	lx.eval('select.drop edge')
	try:
		lx.eval('!!select.useSet %s select' % vnormkit.hardEdgeSetName)
	except RuntimeError:
		lx.eval('noop')

	if len(args) > 1:
		if args[1].lower() ==  'soft':
			lx.eval('select.edgeInvert')
	
	if selRestore[1]:
		lx.eval('!!select.deleteSet %s' % vnormkit.tempSetEdge)
		selRestore[1] = False
	selectionType = 'edge'
#______________________________________________________________________________________________________ Update Smoothing
elif args[0].lower() == 'update':
	updateSmoothing()
	dropSelection()
#________________________________________________________________________________________________________________ Harden
elif args[0].lower() == 'harden':
	#lx.out('Hardening.')
	if len(args) > 1:
		if args[1].lower() == 'all':
			#lx.out('All.')
			lx.eval('select.drop edge')
			lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)
		elif args[1].lower() == 'uv':
			#lx.out('UV Borders.')
			lx.eval('select.drop edge')
			checkUVSelection()
			lx.eval('!!uv.selectBorder')
			if lx.eval1('query layerservice edge.N ? selected') > 0:
				lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)
		elif args[1].lower() == 'polyborder':
			#lx.out('Polygon Borders.')
			lx.eval('select.drop edge')
			lx.eval('select.type polygon')
			lx.eval('select.boundary')
			if lx.eval1('query layerservice edge.N ? selected') > 0:
				lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)
		elif args[1].lower() == 'material':
			hardEdgesByMaterial()
			smoothMaterials()
		elif args[1].lower() == 'angle':
			if len(args) > 2:
				if args[2].lower() == 'softennonhard':
					lx.eval('select.drop edge')
					lx.eval('select.editSet %s remove \"\"' % vnormkit.hardEdgeSetName)
			angle = lx.eval1("user.value vnormkit.angle ?")
			angle = max(min(angle, 180.0), 0.0)
			hardEdgesByAngle(angle)
			smoothMaterials()
	elif lx.eval1('query layerservice edge.N ? selected') > 0:
		#lx.out('Selection.')
		lx.eval('select.editSet %s add \"\"' % vnormkit.hardEdgeSetName)
	updateSmoothing()
	dropSelection()
#________________________________________________________________________________________________________________ Soften
elif args[0].lower() == 'soften':
	#lx.out('Softening.')
	if len(args) > 1:
		if args[1].lower() == 'all':
			#lx.out('All.')
			lx.eval('select.drop edge')
			lx.eval('select.editSet %s remove \"\"' % vnormkit.hardEdgeSetName)
		elif args[1].lower() == 'polyborder':
			#lx.out('Polygon Borders.')
			lx.eval('select.drop edge')
			lx.eval('select.type polygon')
			lx.eval('select.boundary')
			if lx.eval1('query layerservice edge.N ? selected') > 0:
				lx.eval('select.editSet %s remove \"\"' % vnormkit.hardEdgeSetName)
	elif lx.eval1('query layerservice edge.N ? selected') > 0:
		#lx.out('Selection.')
		lx.eval('select.editSet %s remove \"\"' % vnormkit.hardEdgeSetName)
	updateSmoothing()
	dropSelection()

# Restore original selection. 
vnormkit.restoreSelection(selRestore)

vnormkit.renameNormalMapsCustom(True)

# Return to original selection type.
lx.eval('select.type %s' % selectionType)