#python

import lx
from math import sqrt

# Helper functions and common values that are shared across my kits.


#_______________________________________________________________________________________________________________________ Variables

defaultNormalName = 'vert_normals'

hardEdgeSetName = 'HardEdges'
openEdgeSetName = 'OpenEdges'
smoothBoundsSetName = 'SmoothBoundsEdges'

tempSetVert = 'vertPrevSelection'
tempSetEdge = 'edgePrevSelection'
tempSetPoly = 'polyPrevSelection'

#_______________________________________________________________________________________________________________________ Vector Maths

# Return dot product of two vectors.
def vector_dot(vecA, vecB):
	dot = 0.0
	for a, b in zip(vecA, vecB):
		dot += a * b
	return max(-1.0, min(dot, 1.0))

# Get vector length.
def vector_length(vec):
	return sqrt(vec[0]**2 + vec[1]**2 + vec[2]**2)

# Get normalized vector.
def vector_normalize(vec):
	length = vector_length(vec)
	if length != 0.0:
		return [x / length for x in vec]
	else:
		return [0.0] * 3

# Get normalized vector with assigned length.
# DOES NOT CHECK FOR ZERO LENGTH, BE SURE YOU'RE PASSING A NON-ZERO VALUE.
def vector_normalize_length(vec, length):
	return [x / length for x in vec]

# Multiply vector A with scalar B.
def vector_multiply_scalar(vec, scalar):
	return [x * scalar for x in vec]

# Subtract vector B from vector A.
def vector_subtract(vecA, vecB):
	return [a - b for a, b in zip(vecA, vecB)]

# Add vector B to vector A.
def vector_add(vecA, vecB):
	return [a + b for a, b in zip(vecA, vecB)]

# Get the area of the triangle with points point A, B and C.
def triangle_area(pointA, pointB, pointC):
	edgeA = vector_subtract(pointB, pointA)
	edgeB = vector_subtract(pointC, pointB)
	edgeC = vector_subtract(pointA, pointC)

	sideA = vector_length(edgeA)
	sideB = vector_length(edgeB)
	sideC = vector_length(edgeC)

	perim = (sideA + sideB + sideC) / 2

	return sqrt(perim * (perim - sideA) * (perim - sideB) * (perim - sideC))

#_______________________________________________________________________________________________________________________ Selection Type, Save/Restore

# Get selection type.
def getSelectionType():
	selType = lx.eval1('query layerservice selmode ?')
	if selType == 'vertex' or selType == 'edge' or selType == 'polygon':
		return selType
	else:
	 	return 'item'

# Create temporary selection sets for selected elements, if there are any.
def saveSelection():
	result = [False] * 3

	layers = lx.evalN('query layerservice layers ? fg')
	for layer in layers:
		layerID = lx.eval1('query layerservice layer.ID ? %s' % layer)
		if lx.eval1('query layerservice vert.N ? selected') > 0:
			result[0] = True
		if lx.eval1('query layerservice edge.N ? selected') > 0:
			result[1] = True
		if lx.eval1('query layerservice poly.N ? selected') > 0:
			result[2] = True
		
	if result[0]:
		lx.eval('select.type vertex')
		lx.eval('select.editSet %s add' % tempSetVert)
	if result[1]:
		lx.eval('select.type edge')
		lx.eval('select.editSet %s add' % tempSetEdge)
	if result[2]:
		lx.eval('select.type polygon')
		lx.eval('select.editSet %s add' % tempSetPoly)

	return result

# Select then remove selection set for previously selected elements.
def restoreSelection(selRestore):
	is601 = False
	if lx.eval1('query platformservice appversion ?') < 701:
		is601 = True

	if selRestore[0]:
		lx.eval('select.type vertex')
		lx.eval('!!select.useSet %s select' % tempSetVert)
		if is601:
			lx.eval('!!select.vertexMap %s pick replace' % tempSetVert)
			lx.eval('!!vertMap.delete pick')
		else:
			lx.eval('!!vertMap.deleteByName PICK %s' % tempSetVert)
	if selRestore[1]:
		lx.eval('select.type edge')
		lx.eval('!!select.useSet %s select' % tempSetEdge)
		if is601:
			lx.eval('!!select.vertexMap %s epck replace' % tempSetEdge)
			lx.eval('!!vertMap.delete epck')
		else:
			lx.eval('!!vertMap.deleteByName EPCK %s' % tempSetEdge)
	if selRestore[2]:
		lx.eval('select.type polygon')
		lx.eval('!!select.useSet %s select' % tempSetPoly)
		if is601:
			lx.eval('!!select.deleteSet %s' % tempSetPoly)
		else:
			lx.eval('!!poly.renameTag %s {} pick' % tempSetPoly)

#_______________________________________________________________________________________________________________________ VertMap Helpers

def doesNormalMapExist(name):
	normalMaps = lx.evalN('query layerservice vmaps ? normal')
	normalMapsNamed = [i for i in normalMaps if lx.eval1('query layerservice vmap.name ? %s' % i) == name]
	return len(normalMapsNamed) > 0

def doesEdgeSetExist(name):
	epickMaps = lx.evalN('query layerservice vmaps ? all')
	epickMapsNamed = [i for i in epickMaps if lx.eval1('query layerservice vmap.type ? %s' % i) == 'edgepick' and lx.eval1('query layerservice vmap.name ? %s' % i) == name]
	return len(epickMapsNamed) > 0

#_______________________________________________________________________________________________________________________ Open Edge Helpers

# Save all open edges to a selection set.
def saveOpenEdges():
	lx.eval('select.type edge')
	lx.eval('select.drop edge')
	lx.eval('select.boundary')
	if lx.eval1('query layerservice edge.N ? selected') > 0:
		# Yes, we do.
		lx.eval('select.editSet %s add' % openEdgeSetName)
	lx.eval('select.drop edge')

# Weld all open edges, split previously open ones.
def reweldAndSplit():
	# Weld all open edges.
	lx.eval('select.type edge')
	lx.eval('select.drop edge')
	lx.eval('select.boundary')
	lx.eval('select.convert vertex')
	lx.eval('tool.set vert.merge on')
	lx.eval('tool.setAttr vert.merge dist 0.0')
	lx.eval('tool.doApply')
	lx.eval('tool.set vert.merge flush 0')
	lx.eval('select.drop vertex')

	# Split he previously open edges, if there were any.
	lx.eval('select.type edge')

	# try:
	# 	lx.eval('!!select.vertexMap %s epck replace' % openEdgeSetName)
	# except RuntimeError:
	# 	lx.eval('noop')
	# else:
	# 	lx.eval('select.drop edge')
	# 	lx.eval('!!select.useSet %s select' % openEdgeSetName)
	# 	if lx.eval1('query layerservice edge.N ? selected') > 0:
	# 		lx.eval('edge.split')
	# 	lx.eval('vertMap.deleteByName EPCK %s' % openEdgeSetName)

	if doesEdgeSetExist(openEdgeSetName):
		lx.eval('select.drop edge')
	 	lx.eval('!!select.useSet %s select' % openEdgeSetName)
	 	if lx.eval1('query layerservice edge.N ? selected') > 0:
	 		lx.eval('edge.split')

	 	if lx.eval1('query platformservice appversion ?') < 701:
			lx.eval('!!select.vertexMap %s epck replace' % openEdgeSetName)
			lx.eval('!!vertMap.delete epck')
		else:
			lx.eval('!!vertMap.deleteByName EPCK %s' % openEdgeSetName)

#_______________________________________________________________________________________________________________________ MODO Version Check

def isSP1():
	if lx.eval1('query platformservice appbuild ?') == 59358:
		return True
	else:
		return False

#_______________________________________________________________________________________________________________________ User Value Sanity Checkers

def checkNormalMapName():
	userNormalName = lx.eval1('user.value vnormkit.normname ?')
	userNormalName = ''.join(userNormalName.split())
	if len(userNormalName) == 0:
		lx.eval('user.value vnormkit.normname %s' % defaultNormalName)


def checkSmoothingAngle():
	smoothingAngle = lx.eval1('user.value vnormkit.angle ?')
	smoothingAngle = max(min(smoothingAngle, 180.0), 0.0)
	lx.eval('user.value vnormkit.angle %s' % smoothingAngle)

#_______________________________________________________________________________________________________________________ Normal Map Renamers

def renameNormalMaps(normalName):
	normalMaps = lx.evalN('query layerservice vmaps ? normal')
	numNormalMaps = len(normalMaps)
	if numNormalMaps > 0:
		tempNormalName = lx.eval1('query layerservice vmap.name ? %s' % normalMaps[0])
		while numNormalMaps > 1:
			normalMaps = lx.evalN('query layerservice vmaps ? normal')
			numNormalMaps = len(normalMaps)
			normalMapName = lx.eval1('query layerservice vmap.name ? %s' % normalMaps[1])
			lx.eval('!!select.vertexMap {%s} norm replace' % normalMapName)
			lx.eval('!!vertMap.name {%s} norm active' % tempNormalName)
			normalMaps = lx.evalN('query layerservice vmaps ? normal')
			numNormalMaps = len(normalMaps)

		tempNormalName = lx.eval1('query layerservice vmap.name ? %s' % normalMaps[0])
		if tempNormalName != normalName:
			lx.eval('!!select.vertexMap {%s} norm replace' % tempNormalName)
			lx.eval('!!vertMap.name {%s} norm active' % normalName)

def renameNormalMapsDefault():
	renameNormalMaps(defaultNormalName)

def renameNormalMapsCustom(check):
	doIt = lx.eval1("user.value vnormkit.usenormname ?")
	if not check:
		doIt = True

	if doIt:
		normalName = lx.eval1("user.value vnormkit.normname ?")
		renameNormalMaps(normalName)